#!/bin/bash
shopt -s extglob  # Shell options: turn on advanced globbing (http://www.linuxjournal.com/content/bash-extended-globbing)

################################################################################
#  Instructions:
#  Clone the RNAstructure source from Git into a local folder: 
#    $ url=git@rna-git.urmc-sh.rochester.edu:mathews-lab/RNAstructure.git
#  	 $ git clone --depth 1 --branch master $url
#         
#  Run the build script from inside RNAstructure/RELEASE/
#  	 $ cd RNAstructure/RELEASE
#    $ bash build;
#  This will create a BUILD_<VERSION> directory at the same folder-level as RNAstructure:
#  	<PARENT_DIR>
#		RNAstructure
#  			RELEASE
#  			exe
#  			...
#  		BUILD_v5.9
#  			RNAstructureSource
#  			RNAstructure_x86
#  			...
#  You can change the output folder, version, build-date and several other 
#  settings using flags. run `bash build -h` to show usage instructions.
################################################################################

# Some global settings
SCRIPT_NAME=buildRelease  # The "friendly" name of this script.
PROJ_NAME=RNAstructure  # Project Name (used in various places)
TAR_OWNER=$PROJ_NAME TAR_GROUP=$PROJ_NAME  # Specify the owner and group for tar archives.

# GUI_INCLUDE - a list of files and folders from RNAstructure that should be included in GUI releases. (See copyEx function for details about format.
GUI_INCLUDE=(data_tables examples gpl.txt Version.txt ReadMe.txt manual/GUI=manual --dir=exe \$JNI_LIB \*.jar \*.bat \*Script )

PREV_DIR="$PWD"
SCRIPTS_DIR=$(dirname "${BASH_SOURCE[0]}")
BASE_DIR=$(cd "$SCRIPTS_DIR/.." && pwd)
#SCRIPTS_DIR="$BASE_DIR/scripts"
RES_DIR="$BASE_DIR/resources"

# Include Utility functions
source $SCRIPTS_DIR/utils.bash
# Include Release-specific functions
source $SCRIPTS_DIR/releaseUtils.bash

DEF_PROMPT=1 # By default, show a confirmation prompt
DEF_SRC_DIR=$(cd "$BASE_DIR/.."; pwd)        # The directory that contains the raw RNAstructure source folder (e.g. cloned from git.)
DEF_OUT_DIR=$(cd "$BASE_DIR/../.."; echo "$PWD/BUILD_v[VERSION]") # The location where the output of the release build should be stored.
DEF_ARCH="$HOSTTYPE"
OS=$(getOS)

ALL_TASKS='c,b,t,j,a'
DEF_TASKS='b,j,a'

BUILD_TYPE=t
TASKS="$DEF_TASKS"

# Initialize current Release configuration.
source $BASE_DIR/release.cfg

# Store VERSION and RELEASE_DATE values set in 'current_release_config.sh'  These are used only for display in the showusage function.
: ${DEF_VERSION:=$VERSION} ${DEF_RELEASE_DATE:=$RELEASE_DATE}

showusage() {
	local N=$'\n'
	local I=$N'                   '
#23456789|123456789|123456789|123456789|123456789|123456789|123456789|123456789|  <-- 80 characters
	echo -n \
"$SCRIPT_NAME [OPTIONS...]
 -b=BUILD     
      Specify the build type: t=Text, s=Source, g=GUI.  (Default: t)
 -v=VERSION
       The current version of the release. (Default: $DEF_VERSION)
 -d=RELEASE_DATE
       The expected release date.  (Default: $DEF_RELEASE_DATE)
 -a=TARGET_ARCH
 	   The target machine architecture (e.g. i686, x86_64). (Default: $HOSTTYPE)
 -s=SRC_DIR
       The RNAstructure source directory, checked out from Git.
       (Default: '$DEF_SRC_DIR')
 -o=OUT_DIR
       The output directory under which temporary build folders and
       final release packages should be saved. 
       (Default: '$DEF_OUT_DIR')
 -t=TASKS | t=[+ADD],[-REMOVE]
       Specify which task(s) should be performed. COMMON TASKS include:  
         c=clean, b=build-text, t=test, j=build-java, a=archive. 
       The default set of tasks is '$DEF_TASKS'.  Tasks can be added or removed from
         this default list using the syntax -t=[+A,B...],[-Y,Z...]  If the list
         of tasks is not prefixed (with + or -) it is assumed to be specified
         exactly--replacing the default list. So '-t=c,b,t,a' AND '-t=+c,t,-j' 
         both describe the same list of tasks: c,b,t,a.
 -p|--prompt 
       Show information about the build and prompt the user to continue before
         proceeding. (This is the default unless -f is given)
 -f|--force 
       Skip the default confirmation prompt (see --prompt)
 -h|--help
       Show this help message.
";}

parseArgs() {
	local arg opt val argNum=0 ordArgNum=0 taskFlags
	for arg in "$@"; do
		((argNum++))
		if [[ $arg == -* ]]; then
			opt="${arg%%[=:]*}"
			val="${arg:${#opt}+1}"
			case "$opt" in 
				-v) VERSION="$val";;
				-a) ARCH="$val";;
				-d) RELEASE_DATE="$val";;
				-b) BUILD_TYPE="$val";;
				-t) taskFlags+=" $val";;
				-s) SRC_DIR="$val"; ;; # readVer;;
				-o) OUT_DIR="$val";;
				-h|--help) showusage; exit 0 ;;
				-f|--force) DEF_PROMPT= ;;
				-p|--prompt) DEF_PROMPT=1 ;;
				*) abort -U 'Invalid argument (%d): "%s"' $argNum "$arg" ;;
			esac
		else
			case $((++ordArgNum)) in 
				1) BUILD_TYPE="$arg";;
				2) VERSION="$arg";;
				3) ARCH="$arg";;
				4) RELEASE_DATE="$arg";;
				*) abort -U 'Invalid argument (%d): "%s"' $argNum "$arg" ;;
			esac
		fi
	done
	TASKS=$(parseTasks "$TASKS" "$taskFlags")
}

parseArgs "$@"

# Set values for parameters the user did not specify.
: ${SRC_DIR:=$DEF_SRC_DIR} ${OUT_DIR:=$DEF_OUT_DIR} ${VERSION:=$DEF_VERSION} ${ARCH:=$DEF_ARCH} ${RELEASE_DATE:=$DEF_RELEASE_DATE}
SRC_DIR=$(replaceVars "$SRC_DIR")
OUT_DIR=$(replaceVars "$OUT_DIR")

if [[ $ARCH == *64 ]]; then BITS=64; else BITS=32; fi

[[ $BUILD_TYPE ]] || 	abort -U 'Please specify the build type: t=text, s=source, or g=GUI.'
[[ $VERSION ]] || 		abort -U 'Please specify the release version, e.g. "5.8".'
[[ $RELEASE_DATE ]] || 	abort -U 'Please specify the release date, e.g. "2015-10-25".'
[[ $OS && $ARCH ]] || 	abort -U 'The OS and architcture could not be determined. Please specify target architecture.'
#[[ $SRC_DIR ]] || 	abort -U 'The specified root directory is invalid. (SRC_DIR="%s").' "$SRC_DIR"

export RNA_MAKE_ARCH=$ARCH
export RNA_MAKE_OS=$OS

MAKE_TARGET=all

NAME=RNAstructure
os=$(tolower "$OS")
WDIR_Text=$OUT_DIR/${os}$BITS
WDIR_Gui=$OUT_DIR/${os}GUI
WDIR_Source=$OUT_DIR/RNAstructureSource
ARC_EXT=tgz

case $OS in 
	Windows) JNI_LIB=RNAstructure_GUI_[BITS].dll
             ARC_EXT=zip
             ;;
	Linux) 	 JNI_LIB=libRNAstructure_GUI_[BITS].so
			 ;;
	Mac) 	 JNI_LIB=libRNAstructure_GUI.dylib 		
             MAKE_TARGET=serial
             ;;
	*) abort 'Unknown OS (%s)- cannot determine JNI library name.' "$OS" ;;
esac

JNI_LIB1=${JNI_LIB/"[BITS]"/32}
JNI_LIB2=${JNI_LIB/"[BITS]"/64}
JNI_LIB=${JNI_LIB/"[BITS]"/$BITS}
LIB_OUT_DIR=${OUT_DIR}/lib

cat <<EOF
  VERSION: $VERSION  RELEASE_DATE: $RELEASE_DATE   OS=$OS  ARCH=$ARCH ($BITS-bit)
  BUILD_TYPE: $BUILD_TYPE (t=Text, s=Source, g=GUI)
  BASE DIR:   $BASE_DIR
  SOURCE DIR: ${SRC_DIR/#$BASE_DIR/'<BASE_DIR>'}
  OUTPUT DIR: ${OUT_DIR/#$BASE_DIR/'<BASE_DIR>'}
  TASKS: $TASKS (c=clean, b=build-text, t=test, j=build-java, a=archive)
EOF
[[ $DEF_PROMPT -ne 0 ]] && read -n1 -r -p "Press any key to continue, or Ctrl+C to quit..."

mkdir -p "$OUT_DIR" || abort 'Could not create release output directory: "%s"' "$OUT_DIR"
#mkdir -p "$SRC_DIR" || abort 'Could not create source directory: "%s"' "$SRC_DIR"

# if [[ -f $SRC_DIR/CVS/Entries ]]; then
# 	cd $SRC_DIR
# 	cvs -q update -dP || abort 'Failed to update RNAstructure from Git.'
# else
# 	cd $SRC_DIR/..
# 	cvs -q co -d $(basename "$SRC_DIR") -P RNAstructure || abort 'Failed to checkout RNAstructure from CVS.'
# fi

#CMD_VERSION=$(sed -n 's/\s*version\s*=\s*\([0-9.]\+\)\s*;/\1/ip'  src/ParseCommandLine.cpp)
#[[ $CMD_VERSION == $VERSION ]] || abort "Version mismatch: Version=$VERSION, ParseCommandLine=$CMD_VERSION"

set -f # Turn off globbing, so the patterns below do not need to be quoted.
TAR_EXCLUDE=(
 	*.o 						 # object files
	./tests/*_OUTPUT 			 # regression test output
	./exe/java ./exe/java_tests  # java compilation output folders
	./exe/*.log 				 # log files from java compilation etc.
	*.exclude
	old_*						 # e.g. java_interface/old_java_folders manuals/javadoc/old_package_structure
	*.iws						 # User-specific IntelliJ IDEA configuration files
	./java_interface/installer
	./java_interface/ide/eclipse
	.#*	 *~  #*#  *.bak *.old    # backup/archived/versioned files 
	CVS RCS SCCS .git .gitignore .cvsignore .svn cvs.ignore .gitattributes # VCS files (CVS, git, SVN etc)
); set +f # Turn globbing back on.

SYNC_EXCLUDE=("${TAR_EXCLUDE[@]/#.\//\/}") # replace ./ with / at the beginning of the exclusions because rsync considers patterns to be anchored to the sync root when they start with '/' whereas tar considers patterns anchored with  './'
[[ $OS == Windows ]] || TAR_EXCLUDE+=('*.bat')  # No need for bat files on Linux or Mac (but DO keep shell scripts on Windows for cygwin etc)

# The following function is only relevant inside the subsequent for-loop. 
initBuild() { # parameters: (1) location of WORK_DIR;  (2...) Source files or directories to copy (via rsync)
	WORK_DIR="$1" # e.g. TDIR, SDIR
	mkdir -p $WORK_DIR || abort 'Could not create release output directory: "%s"' "$WORK_DIR"
	RNAS_DIR=$WORK_DIR/$NAME # RNAstructure
	RNAS_GUI_DIR=$RNAS_DIR-GUI

	cd "$SRC_DIR"  # RNAstructure/
	cpsync "${@:2}" "$RNAS_DIR"   # abort on failure is already built-in
	# Write release information (Version, Date, etc) to Version.txt
	writeVer $BUILD "$RNAS_DIR"
	cd "$RNAS_DIR"
}

for BUILD in t g s; do
	[[ $BUILD_TYPE == *$BUILD* ]] || continue
	case $BUILD in 
		s)  # Build Source Archive
			initBuild "$WDIR_Source"  "$SRC_DIR/"   # Defines the working directory and source files to copy into it.
			# No need to build or test any files. Simply create the archive.
			cd $RNAS_DIR/.. # ouside the RNAstructure directory
			[[ $TASKS == *a* ]] && {
				buildArchive ${OUT_DIR}/RNAstructureSource.$ARC_EXT $NAME
			}
			;;
		t)  # Build Text Interface executables and Archive
			initBuild "$WDIR_Text"  "$SRC_DIR/" # Defines the working directory and source files to copy into it.
			[[ $TASKS == *c* ]] && {
				cd "$RNAS_DIR"
				make realclean || abort 'Could not make clean.'
				cd java_interface && make realclean || abort 'Could not make java clean.' 
				cd "$RNAS_DIR"
			}
			[[ $TASKS == *b* ]] && {
				rm -f ${WORK_DIR}/build@(Result|Log|Err)
				printf "\nBuilding $NAME...\nSee  %s for progress...\n" ${WORK_DIR}/buildLog
				cd "$RNAS_DIR"
				(make $MAKE_TARGET > ${WORK_DIR}/buildLog; echo -n $? > ${WORK_DIR}/buildResult) 2>&1 | tee ${WORK_DIR}/buildErr		
				EXIT_CODE=$(< ${WORK_DIR}/buildResult)
				[[ $EXIT_CODE -eq '0' ]] || abort 'Make executables failed.'
			}
			[[ $TASKS == *t* ]] && {
				rm -f ${WORK_DIR}/test@(Result|Log|Err)
				cd $RNAS_DIR/tests
				printf "\nTesting $NAME...\nSee  %s for progress.\n" ${WORK_DIR}/testLog
				(make $MAKE_TARGET > ${WORK_DIR}/testLog; echo -n $? > ${WORK_DIR}/testResult) 2>&1 | tee ${TDIR}/testErr
				cd ..
			}
			[[ $TASKS == *j* ]] && {
				rm -f ${WORK_DIR}/jni@(Result|Log|Err)
				rm -f ${WORK_DIR}/$JNI_LIB ${RNAS_DIR}/exe/$JNI_LIB
				cd "$SRC_DIR"
				JNI_FILES=(*.h Makefile java_interface RNA_class src TurboFold )
				cpsync "${JNI_FILES[@]}"  "$RNAS_GUI_DIR"  # abort on failure is already built-in
				cd "$RNAS_GUI_DIR"  
				[[ $TASKS == *c* ]] && { make realclean || abort 'Could not make clean.'; }
				cd java_interface
				[[ $TASKS == *c* ]] && { make realclean || abort 'Could not make JAVA clean.'; }
				printf "\nBuilding $NAME Java Native Interface...\nSee  %s for progress...\n" ${WORK_DIR}/jniLog
				mkdir -p "$RNAS_GUI_DIR/exe"
				(make native > ${WORK_DIR}/jniLog; echo -n $? > ${WORK_DIR}/jniResult ) 2>&1 | tee ${WORK_DIR}/jniErr
				EXIT_CODE=$(< ${WORK_DIR}/jniResult)
				JNI_LIB_PATH=$RNAS_GUI_DIR/exe/$JNI_LIB
				[[ $EXIT_CODE -eq '0' && -f $JNI_LIB_PATH ]] || abort 'Make Java GUI native interface failed.'
				mkdir -p "$LIB_OUT_DIR"
				cp -f $JNI_LIB_PATH "$LIB_OUT_DIR"  # Copy compiled library to 'lib' dir in the output folder
				cp -f $JNI_LIB_PATH ${RNAS_DIR}/exe # Also copy to exe folder in RNAstructure working directory.
				cd "$RNAS_DIR"
			}
			[[ $TASKS == *a* ]] && {
				if [[ $OS == Mac ]]; then
					ARCHIVE_FILE=RNAstructureTextInterfacesMac.tgz
				elif [[ $OS == Windows ]]; then
					# Do not build a tar. Just copy the files required for the Windows installer.
					#tmpDir=${OUT_DIR}/Win$BITS
					#mkdir -p $tmpDir
					#cd $RNAS_DIR
					#rsync  exe data_tables examples Release gpl.txt Version.txt ReadMe.txt ${tmpDir} ||
					#ARCHIVE_FILE= #Do not build
					# This isn't necessary, but it is useful for troubleshooting.
					ARCHIVE_FILE=RNAstructure${OS}TextInterfaces${BITS}bit.zip
				elif [[ $BITS == 32 ]]; then
					ARCHIVE_FILE=RNAstructure${OS}TextInterfaces.tgz
				else
					ARCHIVE_FILE=RNAstructure${OS}TextInterfaces64bit.tgz
				fi
				cd $RNAS_DIR/.. #Outside the RNAstructure directory
				[[ $ARCHIVE_FILE ]] && { buildArchive ${OUT_DIR}/$ARCHIVE_FILE $NAME || abort 'Failed to build tar: "%s"' "${OUT_DIR}/$ARCHIVE_FILE"; }
			}
			;;
		g)  # Build Java GUI and Archive
			# First, see if the native interface already exists
			#[[ -f $OUT_DIR/$JNI_LIB1 ]] || abort 'The Java native interface must be built first. Missing "%s"' "$OUT_DIR/$JNI_LIB1"
			#[[ -f $OUT_DIR/$JNI_LIB2 ]] || abort 'The Java native interface must be built first. Missing "%s"' "$OUT_DIR/$JNI_LIB2"
			GUI_FILES=( Version.txt exe )
			initBuild "$WDIR_Gui"  "${GUI_FILES[@]}"   # Defines the working directory and source files to copy into it.
			BUILD_DIR=$WORK_DIR/BUILD-GUI
			mkdir -p $BUILD_DIR/exe
			cd $SRC_DIR
			cpsync *.h java_interface $BUILD_DIR

			[[ $TASKS == *c* ]] && {
				cd $BUILD_DIR
				make realclean || abort 'Could not make clean.'
				cd java_interface && make realclean || abort 'Could not make java clean.' 
				cd $RNAS_DIR
			}

			[[ $TASKS == *b* ]] && {
				rm -f ${WORK_DIR}/java@(Result|Log|Err)
				printf "\nBuilding $NAME Java GUI...\nSee  %s for progress...\n" ${WORK_DIR}/javaLog
				cd  $BUILD_DIR/java_interface
				(make javaGUI > ${WORK_DIR}/javaLog; echo -n $? > ${WORK_DIR}/javaResult) 2>&1 | tee ${WORK_DIR}/javaErr
				EXIT_CODE=$(< ${WORK_DIR}/javaResult)
				[[ $EXIT_CODE -eq '0' ]] || abort 'Make Java GUI failed.'
				cd $BUILD_DIR/exe
				cp -f *.jar ${RNAS_DIR}/exe || abort 'Failed to copy GUI jar files to output directory (%s).'  "${RNAS_DIR}/exe"
				cd $RNAS_DIR
			}

			# If the above was successful, copy the rest of the files required for the build.
			ERRMSG='The %d-bit Java native interface must be built before continuing. (Missing "%s)"'
			rsync -au $LIB_OUT_DIR/$JNI_LIB  ${RNAS_DIR}/exe 	|| abort "$ERRMSG" $BITS "$LIB_OUT_DIR/$JNI_LIB"
			rsync -au $LIB_OUT_DIR/$JNI_LIB1 ${RNAS_DIR}/exe 	|| abort "$ERRMSG" 32 "$LIB_OUT_DIR/$JNI_LIB1"
			rsync -au $LIB_OUT_DIR/$JNI_LIB2 ${RNAS_DIR}/exe 	|| abort "$ERRMSG" 64 "$LIB_OUT_DIR/$JNI_LIB2"
			cd $SRC_DIR
			cpsync manual/GUI/ $RNAS_DIR/manual  # abort on failure is already built-in
			cpsync data_tables examples gpl.txt ReadMe.txt ${RNAS_DIR}
			cpsync java_interface/scripts/RNAstructureScript ${RNAS_DIR}/exe
			cpsync java_interface/ui-resources/logo.png ${RNAS_DIR}
			[[ $OS == Windows ]] && cpsync java_interface/scripts/*.bat ${RNAS_DIR}/exe
			chmod 755 ${RNAS_DIR}/exe/RNAstructureScript  ${RNAS_DIR}/exe/*.jar ${RNAS_DIR}/exe/$JNI_LIB1 ${RNAS_DIR}/exe/$JNI_LIB2

			# tmpDir=${OUT_DIR}/$OS-GUI/RNAstructure
			# mkdir -p ${tmpDir}/exe || abort 'Failed to create temporary GUI directory: "%s"' "${tmpDir}/exe"
			# (   GUI_ITEMS=''
			# 	cd $RNAS_DIR   							|| exit 101
			# 	cp -rf  $GUI_ITEMS $tmpDir    			|| exit 102
			#   	     || exit 103
			# 	cd exe
			# 	#shopt -s nullglob extglob
			# 	cp -f *.jar *.bat *Script ${tmpDir}/exe 	|| exit 104
			# 	ERRMSG='The %d-bit Java native interface must be built before continuing. (Missing "%s)"'
			# 	cp  -f $LIB_OUT_DIR/$JNI_LIB  ${tmpDir}/exe 	|| abort -e=105 "$ERRMSG" $BITS "$LIB_OUT_DIR/$JNI_LIB"
			# 	[[ $JNI_LIB1 == $JNI_LIB ]] || { cp  -f $LIB_OUT_DIR/$JNI_LIB1  ${tmpDir}/exe 	|| abort -e=106 "$ERRMSG" 32 "$LIB_OUT_DIR/$JNI_LIB1"; }
			# 	[[ $JNI_LIB2 == $JNI_LIB ]] || { cp  -f $LIB_OUT_DIR/$JNI_LIB2  ${tmpDir}/exe 	|| abort -e=107 "$ERRMSG" 64 "$LIB_OUT_DIR/$JNI_LIB2"; }

			# ) || abort -e=$? 'Failed to copy GUI files or directories.'

			[[ $TASKS == *a* ]] && {
				if [[ $OS == Mac ]]; then
					bash "$SCRIPTS_DIR"/buildMacBundle.sh -B
					#ARCHIVE_FILE=RNAstructureForMac.tgz
					#BUNDLE_DIR=${RES_DIR}/resources/MacBundle
					##CP='rsync -auv'
					#APP_DIR=$OUT_DIR/RNAstructure.app
					#CNT_DIR=$APP_DIR/Contents
					#echo "Building the Mac bundle: " "$APP_DIR"
					#cpsync $BUNDLE_DIR/ $APP_DIR
					#cpsync ${RNAS_DIR}/exe/$JNI_LIB $CNT_DIR/MacOS
					#cpsync ${RNAS_DIR}/exe/RNAstructure.jar $CNT_DIR/Java
					#cpsync $SRC_DIR/data_tables $CNT_DIR/Resources
					#JDK_DIR='jdk1.8.0_60.jdk'
					#replaceVars < "$RES_DIR/Info.plist.tpl" > "$CNT_DIR/Info.plist"  || abort 'Failed to write Info.plist'
				else
				#elif [[ $BITS == 32 ]]; then
					ARCHIVE_FILE=RNAstructureFor${OS}.tgz
				#else
				#	ARCHIVE_FILE=RNAstructureFor${OS}64bit.tgz
				fi
				cd $RNAS_DIR/..  #Outside the RNAstructure directory
				[[ $ARCHIVE_FILE ]] && { buildArchive ${OUT_DIR}/$ARCHIVE_FILE $NAME || abort 'Failed to build tar: "%s"' "${OUT_DIR}/$ARCHIVE_FILE"; }
			}
			;;
	esac
done

echo Build Complete

# make clean
# shopt -s nullglob extglob
# cp -t $OUT_DIR  "$TDIR"/exe/@(*.so|*.dll|*.dylib)
# shopt -u nullglob
# cd .. # to TDIR
# make clean
# ( cd exe && rm -rf java/ java_test/ *.jar.log )
# [[ $OS == Windows ]] || rm -f $TDIR/exe/*.bat
# find $TDIR -depth -name CVS | xargs rm -rf #delete CVS folders.
# find $TDIR -name .cvsignore -delete
# rm -rf $TDIR/java_interface/old_*/ \
#     $TDIR/tests/*_OUTPUT/ \
#     $TDIR/manual/javadoc/old_package_structure/
# cd $OUT_DIR
# tar -zvcf RNAstructureLinuxTextInterfaces${BITS}bit.tgz $SNAME

# #Copy GUI
# mkdir -p "$GDIR"
# cd $TDIR
# make realclean #cleans TDIR
# cd java_interface  &&  make GUI  &&  make clean
# cd .. # to TDIR
# (cd exe && rm -rf java/ java_test/ *.jar.log)
# cp -r data_tables examples exe Version.txt manual/GUI $GDIR/
# mv $GDIR/GUI $GDIR/manual
# cd $OUT_DIR
# tar -zvcf RNAstructureLinux${BITS}bit.tgz $GNAME

