/*
 * (c) 2009 Mathews Lab, University of Rochester Medical Center.
 *
 * This software is part of a group specifically designed for the RNAstructure
 * secondary structure prediction program.
 */

package ur_rna.RNAstructureUI;

import ur_rna.RNAstructureUI.utilities.*;
import ur_rna.Utilities.*;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.nio.file.*;
import java.util.prefs.BackingStoreException;
import java.util.prefs.InvalidPreferencesFormatException;
import java.util.prefs.Preferences;

import static ur_rna.Utilities.ObjTools.asBool;
import static ur_rna.Utilities.ObjTools.isEmpty;

/**
 * A class that starts the RNAstructure interface.
 *
 * @author Jessica S. Reuter
 * @author Richard M. Watson
 */
public class RNAstructure
	implements Serializable {
	private static final long serialVersionUID = 20120802;
	public static final AppLog log = AppLog.getDefault();
	public static final String VERSION = "5.8.1";
	public static final String REQUIRED_DLL_VERSION = "5.8.1";
	public static final String YEAR =  "2016";
	public static MRUFileStorage MRUFiles = new MRUFileStorage();
	public static UserOptions options = new UserOptions();
	public static String preferencesNode = "default";

	/**
	 * The main method.
	 *
	 * @param args   The command line arguments (ignored).
	 */
	public static void main( String[] args ) {
		try {
			parseCommandArgs(args); //checks for some command-line arguments and sets System properties.
			log.readSystemProperties(); //update values based on command-line arguments.

			// If the OS is a type of Mac, certain things need to be set in
			// order to make the native Java more "Mac-like."
			if (options.useNativeLAF)
				OSInfo.applyNativeLookAndFeel();

			if (options.useNativeMenus)
				OSInfo.useNativeMenus();

			log.debug("OS: %s Arch: %s Bits: %s%n", System.getProperty("os.name"), System.getProperty("os.arch"), System.getProperty("sun.arch.data.model"));
			log.debug("user.home: " + System.getProperty("user.home"));
			log.debug("user.dir: " + System.getProperty("user.dir"));
			log.debug("CurDir: " + new File(".").getCanonicalPath());
			log.debug("java.library.path: " + System.getProperty("java.library.path"));
			log.debug("java.class.path: " + System.getProperty("java.class.path"));
			if (log.isTraceEnabled())
				System.getProperties().list(log.getTrStream());
			log.debug("Headless: " + java.awt.GraphicsEnvironment.isHeadless());

			if (!loadNativeLib()) {
                System.exit(2);
                return;
            }

			for (String s : ClassLoaderInfo.getMyLoadedLibraries()) {
				final String find = "RNAstructure".toLowerCase();
				String lc = s.toLowerCase();
				if (lc.contains(find))
					log.debug("Loaded Native DLL: " + s);
			}

			options.load(getPrefs().node("options"));
			Resources.verify();
			verifyDataPath();
			StandardFileChooser.loadRecentPaths(getPrefs().node(StandardFileChooser.PREFS));
			setDefaultFileOpenDir();
			MRUFiles.loadFromStorage();

			// The splash screen window is loaded by a native Java library
			// before the JVM loads (based on the corresponding entry
			// in the Manifest file). It is closed automatically as soon
			// as the first window is displayed by Swing/AWT.
			// So there is no need to manipulate the splash screen directly.
			// If running directly, without compiling to JAR you can specify
			// the splash image by passing the flag  -splash:<path_to_image.gif>
			// on the java command line.

			// Get a reference to the splash screen if it has been shown.
			// getSplashScreen will return null if the splash image is not specified in the
			//   manifest or passed in on the command-line.
			// This would happen for example, if the application is launched
			//   in an IDE directly, as opposed to via a JAR file.
			java.awt.SplashScreen splash = java.awt.SplashScreen.getSplashScreen();
			//Note: Splash screen fails on mac app-bundle: https://bugs.openjdk.java.net/browse/JDK-8090606
			if (splash != null && !silentMode())
				Thread.sleep(2000); //give the splash screen 2 seconds to display before launching the main window.

			// Start the GUI.
			new AppMainFrame();
			//JFrame frm = new AppMainFrame();
//			frm.getRootPane().addPropertyChangeListener(new PropertyChangeListener() {
//				@Override
//				public void propertyChange(final PropertyChangeEvent evt) {
//					log.debug("PropertyChangeEvent: " + evt.toString());
//				}
//			});
		} catch( Exception e ) {
			log.error("error loading RNAstructure.", e);
			try {
				java.io.FileWriter w = new java.io.FileWriter("RNAstructure_error.log");
				w.write(DateHelper.getFormattedDate(DateHelper.ISO8601) + "\n");
				e.printStackTrace(new java.io.PrintWriter(w));
				w.flush();
				w.close();
			} catch (Throwable ex) {
				showWarning("Could not write error log:\n" + AppLog.getErrorInfo(ex));
			}
			System.exit( 10 );
		}
	}
	private static boolean loadNativeLib() {
		// Detect the architecture of the Java platform (32 or 64 bits
		// -- this is NOT necessarily the architecture of the OS!!)
		String libExt, jvmBits = System.getProperty("sun.arch.data.model");
		if (jvmBits.equals("64") || jvmBits.equals("32"))
			libExt = "_"+jvmBits;
		else
			libExt = "";

		String[] libNames = { "RNAstructure_GUI" + libExt, "RNAstructure_GUI" };
		Throwable[] errors = new Throwable[libNames.length];
		String details[] = new String[libNames.length];

		if (tryLoadLibrary(libNames, errors)) return true;

		Throwable loadError = null;
		// Inspect and/or modify the exception messages. Make note of any errors that are not simply due to the library not being found.
		for(int i = 0; i < errors.length; i++) {
			details[i] = errors[i] == null ? "Unknown Error" : errors[i].getMessage();
			if (details[i].equals("no " + libNames[i] + " in java.library.path"))
				details[i] = "Not found.";
			else if (loadError == null)
				loadError = errors[i];
		}
		StringBuilder sb = new StringBuilder();
		if (loadError == null) {
			// no real error occurred. Just couldn't find the library.
			sb.append("The required RNAstructure library is missing.\nIt should be in the same directory as the RNAstructure application, and it should be named:\n    ");
			for (int i = 0; i < libNames.length; i++) {
				if (i != 0)
					sb.append("  or  ");
				sb.append(System.mapLibraryName(libNames[i]));
			}
			sb.append("\n\nFor troubleshooting purposes, note that the following locations were searched:\n  ")
					.append(System.getProperty("java.library.path").replace(File.pathSeparator, "\n  "));
		} else {
			sb.append("Unable to load the RNAstructure library!\n")
				.append(loadError.getMessage())
				.append("\n\n-------- Error Details (For all attempted paths): --------\n");

			for(int i = 0; i < errors.length; i++) {
				sb.append(i + 1).append(")  Attempted to load ").append(System.mapLibraryName(libNames[i]))
						.append("\n      Result: ").append(details[i]).append("\n");
			}
			sb.append("\nAttempted Paths:\n").append(System.getProperty("java.library.path").replace(File.pathSeparator, ", "));
		}
		showWarning(sb.toString());
		return false;
	}
	public static Preferences getPrefs() {
		return Preferences.userNodeForPackage(RNAstructure.class).node(preferencesNode);
	}
	public static String getPref(String name) { return getPref(name, null); }
	public static String getPref(String name, String defaultValue) {
//		try {
			Preferences p = getPrefs();
			return p.get(name, defaultValue);
			//if (p.nodeExists(name))
//		} catch (BackingStoreException ex) {
//			// return default below.
//			ex.printStackTrace();
//		}
//		return defaultValue;
	}
	public static void setPref(String name, String value) {
		Preferences p = getPrefs();
		p.put(name, value);
		try {
			p.flush();
		} catch (BackingStoreException ex) {
			ex.printStackTrace();
		}
	}

	private static void setDefaultFileOpenDir() {
		File examples = getExamplesDir();
		if (examples != null)
			StandardFileChooser.setDefaultDir(examples);
	}

	public static File getExamplesDir() {
		File examples = StandardFileChooser.getRecentDir("examples", true);
		if (examples == null)
			examples = PathTools.findLocalPath("examples");
		if (examples != null && examples.exists())
			return examples;
		return null;
	}

	public static boolean silentMode() {
		return asBool(System.getProperty("silentLaunch"), false);
	}

	private static void parseCommandArgs(String[] args) {
		try {
			for(int i = 0; i < args.length; i++) {
				switch (args[i].toLowerCase()) {
					case "-d":  // debug.
						System.setProperty("debug", "true");
						System.setProperty("show-gui-info-menu","true");
						break;
					case "-s": System.setProperty("silentLaunch","true"); break;
					case "-v": System.setProperty("verbose", args[++i]);
					case "-sfc": AppMainFrame.setUseSimpleFileChooser(true); break;
					case "-jlaf": options.useNativeLAF = false; break;
					case "-jmnu": options.useNativeMenus = false; break;
					case "-prefs": preferencesNode = args[++i]; break;
					case "-opt": {// override a user option (stored in preferences)
						int pos = args[++i].indexOf('=');
						if (pos == -1)
							throw new InvalidPreferencesFormatException("The -opt flag requires an argument in the form \"name=value\".");
						UserOptions.override(args[i].substring(0, pos), args[i].substring(pos + 1));
					}
					break;
					default:
						log.error("Unknown command-line option: \"%s\" (position: %s).",  args[i], i);
				}
			}
		} catch (Throwable e) {
			log.error("error parsing command line.", e);
		}
	}

	static String findDataPath() {
		String paths[] = { ".", "..", "./resources", "../Resources" };
		String fullPaths[] = { "{path}/", "{path}/data_tables", "{app}/{path}/", "{app}/{path}/data_tables" };

		String appPath = PathTools.getAppPath(RNAstructure.class);
        String filter = "*.dat";
        String realpath;
		for (String path : paths) {
			for (String full : fullPaths) {
                String fullPath = full.replace("{app}", appPath).replace("{path}", path);
                File dir = new File(fullPath);
                //System.out.println("Trying: " + fullPath + " ==> " + dir.getAbsolutePath());
                if (dir.exists())
                    try {
                        realpath = dir.getCanonicalPath();
                        if (PathTools.listFiles(realpath, filter).size() != 0)
                            return realpath;
                    } catch (Exception ex) {
                        //Do nothing. continue the loop.
                    }
            }
		}
		return null;
	}

	static String getFullPath(String path) {
		File f = new File(path);
		if (isEmpty(path))
			return "<EMPTY>";

		if (f.exists())
			try {
				return f.getCanonicalPath();
			} catch(IOException ex) {
				return "ERROR: " + ex.getMessage();
			}
		else
			return "<File not found>";
	}
	static boolean fileExists(String path) { return new File(path).exists(); }

	static boolean verifyDataPath() {
		final String vardef = "The DATAPATH environment variable";
		final String APP_PATH = "{APP_PATH}";

		String path = System.getenv("DATAPATH");
		if (path == null) path = "";
		if (path.contains(APP_PATH)) {
			String appPath = PathTools.getAppPath(RNAstructure.class);
			if (appPath == null) appPath = ".";
			log.debug("Using " + APP_PATH + " => " + appPath);
			log.debug("Before replacement: DATAPATH=" + path);
			path = path.replace(APP_PATH, appPath);
			log.debug("After replacement: DATAPATH=" + path);
			setDataPath(path);
		}
		log.debug(String.format("DATAPATH='%s' (%s)", path, getFullPath(path)));

		if (fileExists(path))
			return true;

		// DATAPATH was either NULL or not found. In either case, search for one in standard directories.
		path = findDataPath();
		if (path == null)
			showWarning("The thermodynamic data tables could not be found.\nPlease set the DATAPATH environment variable to the directory where they are stored.\n\t(e.g. <PATH-TO-RNAstructure>/data_tables).");
		else {
			log.debug("The thermodynamic data tables were located at \"" + path + "\". The DATAPATH environment variable will be set to this location.");
			setDataPath(path);
		}
		return path != null;
	}

	static String arrJoin(Object[] arr, String sep) {
		StringBuilder sbStr = new StringBuilder();
		for (Object o : arr) {
			sbStr.append(o.toString());
			sbStr.append(sep);
		}
		if (sbStr.length() > 0)
			sbStr.setLength(sbStr.length()-sep.length());
		return sbStr.toString();
	}

	static void setDataPath(String path) {
		RNAstructureBackendCalculator.setEnvVar("DATAPATH=" + path);
		System.setProperty("DATAPATH", path);
        log.debug("DATAPATH Verification (setDataPath): " + RNAstructureBackendCalculator.getEnvVar("DATAPATH"));
	}

	static boolean tryLoadLibrary(String[] possibleNames, Throwable errors[]) {
		// Add the program's directory to the PATH. (i.e. the directory containing the JAR file or the
		setLibPath(System.getProperty("java.library.path") + File.pathSeparator + PathTools.findLocalPath("."));

		for (int i = 0; i < possibleNames.length; i++) {
			try {
				System.loadLibrary(possibleNames[i]);
				// Seems successful so far. Try to call a known method to verify.
				try {
					String version = RNAstructureBackendCalculator.getVersion();
					if (!isVersionCompatible(REQUIRED_DLL_VERSION, version))
						throw new UnsatisfiedLinkError(String.format("The library version is %s, but a version compatible with %s is required.", version, REQUIRED_DLL_VERSION));
				} catch (UnsatisfiedLinkError ex) {
					throw new UnsatisfiedLinkError("The RNAstructure_GUI library is corrupt or has the wrong version or binary format.");
				}
				return true;
			} catch (Throwable ex) {
				// Error is usually a UnsatisfiedLinkError, but let's catch all of them so we can show them to the user.
				errors[i] = ex;
			}
		}
		return false;
	}
	private static boolean isVersionCompatible(final String referenceVersion, final String testVersion) {
		String[] ref = referenceVersion.split("\\.");
		String[] test = testVersion.split("\\.");
		int len = Math.min(ref.length, test.length);

		if (len < 2) throw new IllegalArgumentException("A version number must have at least two components.");

		// The major version must match identically.
		if (Integer.parseInt(test[0]) != Integer.parseInt(ref[0])) return false;

		// The minor version of the test must be the same or greater than the reference
		if (Integer.parseInt(test[1]) < Integer.parseInt(ref[1])) return false;

		// If the minor version of the test is greater, then don't compare the third component.
		if (Integer.parseInt(test[1]) > Integer.parseInt(ref[1])) return true;

		// If there is no 3rd component, we get here if the third component is provided and the minor versions are the same
		return len == 2 || Integer.parseInt(test[2]) >= Integer.parseInt(ref[2]);
	}

	static void showWarning(String message) {
		log.error("Warning: " + message);
		if (!silentMode())
			Dialogs.showWarning(message, true);
	}
	static void showInfo(String message) {
		log.info("Information: " + message);
		if (!silentMode())
			Dialogs.showMessage(message);
	}
	public static void ExtractExamples() {
		try {
			final String FILE_NAME = "RNAstructureExamples";
			File src = PathTools.findLocalPath("examples");
			if (src == null) {
				Dialogs.showError("The example files source folder could not be found."); //, but you can download them online and extract them manually
				return;
			}
			String userDir = System.getProperty("user.home");
			String path = StandardFileChooser.getFileName(true, null, "Extract Examples Files",   userDir + File.separator + FILE_NAME, "*", null);
			if (path == null) return;
			File[] list = src.listFiles();
			if (list == null) return;

			File dir = new File(path);
			if (!dir.mkdirs()) throw new IOException("Failed to create requested directory: " + path);
			Path dst = dir.toPath();
			for (File sub : list)
				try {
					Files.copy(sub.toPath(), dst.resolve(sub.getName()) , StandardCopyOption.REPLACE_EXISTING);
				} catch (IOException e) {
					e.printStackTrace();
				}
			StandardFileChooser.setRecentDir("examples", dir);
		} catch (Exception ex) {
			Dialogs.showError("Failed to extract example files: " + ex.toString());
		}
	}
	private static String examplesDir;
	public static boolean isExampleFile(final File f) {
		try {
			if (examplesDir == null) {
				File dir = getExamplesDir();
				if (dir == null)
					examplesDir = "<>"; //cannot exist.
				else
					examplesDir = dir.getCanonicalPath();
			}
			return examplesDir.equals(f.getCanonicalFile().getParent());
		} catch (IOException ex) {
			return false;
		}
	}
	/**
	 * Set the java.library.path System property and ensure that it will be re-evaluated by ClassLoaders
	 * @param javaLibraryPath The new setting for java.library.path
     */
	public static void setLibPath(String javaLibraryPath) {
		System.setProperty("java.library.path", javaLibraryPath);
		try {
			// Set the private static ClassLoader.sys_paths field to null, so that java.library.path is re-evaluated when LoadLibrary is called.
			Field fieldSysPath = ClassLoader.class.getDeclaredField("sys_paths");
			fieldSysPath.setAccessible(true);
			fieldSysPath.set(null, null);
		} catch (Exception ex) {
			//
		}
	}

	/**
	 * Compares two version strings.
	 *
	 * Use this instead of String.compareTo() for a non-lexicographical
	 * comparison that works for version strings. e.g. "1.10".compareTo("1.6").
	 *
	 * Note: It does not work if "1.10" is supposed to be equal to "1.10.0".
	 *
	 * @param str1 a string of ordinal numbers separated by decimal points.
	 * @param str2 a string of ordinal numbers separated by decimal points.
	 * @return The result is a negative integer if str1 is _numerically_ less than str2.
	 *         The result is a positive integer if str1 is _numerically_ greater than str2.
	 *         The result is zero if the strings are _numerically_ equal.
	 */
	public static int versionCompare(String str1, String str2) {
		String[] vals1 = str1.split("\\.");
		String[] vals2 = str2.split("\\.");
		int i = 0;
		// set index to first non-equal ordinal or length of shortest version string
		while (i < vals1.length && i < vals2.length && vals1[i].equals(vals2[i])) {
			i++;
		}
		// compare first non-equal ordinal number
		if (i < vals1.length && i < vals2.length) {
			int diff = Integer.valueOf(vals1[i]).compareTo(Integer.valueOf(vals2[i]));
			return Integer.signum(diff);
		}
		// the strings are equal or one string is a substring of the other
		// e.g. "1.2.3" = "1.2.3" or "1.2.3" < "1.2.3.4"
		return Integer.signum(vals1.length - vals2.length);
	}
}
